#include <stdio.h>
#include <stdlib.h>
#include <phidget22.h>
#include "PhidgetHelperFunctions.h"


/**
* Displays info about the attached phidget channel.
* Fired when a Phidget channel with onAttachHandler registered attaches
*
* @param ph The Phidget channel that fired the attach event
* @param *ctx Context pointer
*/
static void CCONV onAttachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	int32_t serialNumber;
	PhidgetHandle hub;
	int32_t hubPort;
	int32_t channel;

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\n[Attach Event]:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\nAttach Event:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Displays info about the detached phidget channel.
* Fired when a Phidget channel with onDetachHandler registered detaches
*
* @param ph The Phidget channel that fired the detach event
* @param *ctx Context pointer
*/
static void CCONV onDetachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	PhidgetHandle hub;
	int serialNumber;
	int hubPort;
	int channel;

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\nDetach Event:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\nDetach Event:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Writes phidget error info to stderr.
* Fired when a Phidget channel with onErrorHandler registered encounters an error in the library
*
* @param ph The Phidget channel that fired the error event
* @param *ctx Context pointer
* @param errorCode the code associated with the error of enum type Phidget_ErrorEventCode
* @param *errorString string containing the description of the error fired
*/
static void CCONV onErrorHandler(PhidgetHandle phid, void *ctx, Phidget_ErrorEventCode errorCode, const char *errorString) {

	fprintf(stderr, "[Phidget Error Event] -> %s (%d)\n", errorString, errorCode);
}

/**
* Sets the event handlers for Phidget Attach, Phidget Detach, Phidget Error events
*
* @param ph The Phidget channel to add event handlers to
* @return 0 if the operation succeeds, 1 if it fails
*/
int SetAttachDetachError_Handlers(PhidgetHandle ch) {
	PhidgetReturnCode prc;
	printf("\n--------------------------------------\n");
	printf("\nSetting OnAttachHandler...\n");
	prc = Phidget_setOnAttachHandler(ch, onAttachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Attach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnDetachHandler...\n");
	prc = Phidget_setOnDetachHandler(ch, onDetachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Detach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnErrorHandler...\n");
	prc = Phidget_setOnErrorHandler(ch, onErrorHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Error Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	return 0;
}

/**
* Outputs the DigitalInput's most recently detected state change.
* Fired when a DigitalInput channel with onStateChangeHandler registered detects a state change
*
* @param pdih The DigitalInput channel that fired the StateChange event
* @param *ctx Context pointer
* @param state The reported state from the DigitalInput channel
*/
static void CCONV onStateChangeHandler(PhidgetDigitalInputHandle pdih, void *ctx, int state) {

	printf("[State Event] -> State: %d\n", state);
}

/**
* Creates a new instance of a DigitalInput channel.
*
* @param pdih Pointer to the PhidgetDigitalInputHandle channel to create
* @return 0 if the operation succeeds, 1 if it fails
*/
static int CreateDigitalInput(PhidgetDigitalInputHandle *pdih) {
	PhidgetReturnCode prc;

	printf("Creating DigitalInput Channel...\n");
	prc = PhidgetDigitalInput_create(pdih);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Creating DigitalInput: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}

/**
* Sets the event handler for DigitalInput's StateChange event
*
* @param pdih The PhidgetDigitalInputHandle channel to add the event to
* @param fptr The callback function to be called when a StateChange event is fired
* @return 0 if the operation succeeds, 1 if it fails
*/
static int SetStateHandler(PhidgetDigitalInputHandle pdih, PhidgetDigitalInput_OnStateChangeCallback fptr) {
	PhidgetReturnCode prc;

	if (fptr) {
		printf("\n--------------------\n"
			"\n  | State change events contain the most recent state received from the device.\n"
			"  | The linked StateChange function will run as an event every time the state changes.\n"
			"  | StateChange events will only occur when a state change happens, therefore no DataInterval is available.\n");
	}
	printf("\n%s OnStateChangeHandler...\n", (fptr ? "Setting" : "Clearing"));
	printf("\n--------------------\n");
	prc = PhidgetDigitalInput_setOnStateChangeHandler(pdih, fptr, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Setting StateChangeHandler: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}


/**
* Creates, configures, and opens a DigitalInput channel.
* Displays state change events for 10 seconds
* Closes out DigitalInput channel
*
* @return 0 if the program exits successfully, 1 if it exits with errors.
*/
int main() {
	PhidgetDigitalInputHandle ch = NULL;

	/*
	 * Allocate a new Phidget Channel object
	 */

	if (CreateDigitalInput(&ch))
		goto error;

	/*
	 * Set matching parameters to specify which channel to open
	 */

	if (SetSerialNumber((PhidgetHandle)ch))
		goto error;

	if (SetVINTProperties((PhidgetHandle)ch))
		goto error;

	if (SetChannel((PhidgetHandle)ch))
		goto error;

	if (SetupNetwork((PhidgetHandle)ch))
		goto error;

	/*
	 * Add event handlers before calling open so that no events are missed.
	 */

	if (SetAttachDetachError_Handlers((PhidgetHandle)ch))
		goto error;

	if (SetStateHandler(ch, onStateChangeHandler))
		goto error;

	/*
	 * Open the channel with a timeout
	 */

	if (OpenPhidgetChannel_waitForAttach((PhidgetHandle)ch, 5000))
		goto error;

	printf("Sampling data for 10 seconds...\n");

	Sleep(10000);

	/*
	 * Perform clean up and exit
	 */

	if (SetStateHandler(ch, NULL))
		goto error;

	printf("\nDone Sampling...\n");

	printf("Cleaning up...\n");
	ClosePhidgetChannel((PhidgetHandle)ch);
	PhidgetDigitalInput_delete(&ch);
	printf("\nExiting...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 0;

error:
	PhidgetDigitalInput_delete(&ch);
	fprintf(stderr, "\nExiting with error(s)...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 1;
}
